
-- Called at the start of a game
function StartFirstTurn()
   WaitUntilNoActivity = false
   SendMessage("Timer.StartGame")
   StartTurn()
end


function StartTurn()
   done_once_per_turn_functions = false
   SendMessage("Timer.StartHotSeatTimer")
   SendMessage("GameLogic.ActivateNextWorm")
   SendMessage("Weapon.Create")
   SetWind()
   --TurnStarted()
   SendMessage("GameLogic.Turn.Started")
   TurnStarted() -- moved here so TeleportIn camera can override startofturn camera
   RunAILogic()
end


function RunAILogic()
	-- ovverride this function to now have AI.
	SendMessage("AI.PerformDefaultAITurn");
	-- actually called to start the AI turn.
	SendMessage("AI.ExecuteActions");
end

function Timer_HotSeatTimedOut()
   SendMessage("Timer.StartTurn")
end


function Weapon_PoweringUpStart()
   SendMessage("Timer.EndTurn")
   SendMessage("Weapon.DisableWeaponChange")
end


function Weapon_Fired_Start()
   -- Called when powering up finishes or fire button presses
   SendMessage("Timer.EndTurn") 
   SendMessage("Weapon.DisableWeaponChange")
end


function Weapon_Fired_End()
   -- called after the short delay which plays the fire animation after which the missile is actually launched
   SendMessage("Timer.StartRetreatTimer")
   SendMessage("Weapon.Delete")
end


function GameLogic_EndTurn_Immediate()
   -- this called by surrender and skip go so they will work even if the above three
   -- functions have been emptied
   SendMessage("Weapon.Delete")
   SendMessage("Timer.EndTurn") 
   SendMessage("Weapon.DisableWeaponChange")
   EndTurn()
end


function Timer_RetreatTimedOut()
   EndTurn()
end


-- The worm whose turn it is has been damaged
function Worm_Damaged_Current()
   SendMessage("Weapon.Delete")
   SendMessage("Timer.EndRetreatTimer")
   SendMessage("Timer.EndTurn")  
   SendMessage("Weapon.DisableWeaponChange")
   EndTurn()
end


function Timer_TurnTimedOut()
   SendMessage("Weapon.Delete")
   SendMessage("Weapon.DisableWeaponChange")
   EndTurn()
end


-- Everything has come stopped moving
function Timer_PostActivityTimedOut()
   SendMessage("GameLogic.AboutToApplyDamage")
   SendMessage("GameLogic.ApplyDamage")
   CheckActivity()
end
   

function CheckActivity()
   Activity = GetData("ObjectCount.Active")
   if Activity == 0 then
      DoPostActivity()
   else
      WaitUntilNoActivity = true
   end
end

function DoPostActivity()
   if done_once_per_turn_functions == true then
      SendMessage("GameLogic.Turn.Ended") 
      TurnEnded()
   else 
      -- Done only once
      SendMessage("Worm.ApplyPoison")
      SetData("DoubleDamage", 0)
      DoOncePerTurnFunctions()
      done_once_per_turn_functions = true
      CheckActivity() -- in case an earthquake or similar was just started
   end
end


-- override this
function TurnStarted()
end

-- override this
function TurnEnded()
   CheckOneTeamVictory()      
end



-- decides if the game has been won or drawn at the end of every turn
-- Note do this last in the TurnEnded function as it finishes with the StartTurn Message
function CheckOneTeamVictory()
   -- Changed to use alliances rather than teams so 2 teams vs 1 can end correctly
   SendMessage("WormManager.GetActiveAlliances")
   AllianceCount = GetData("AllianceCount")
   if AllianceCount == 0 then
      RoundOver()
      -- all the teams have just snuffed it
      SendMessage("GameLogic.Draw")
   elseif AllianceCount == 1 then
      RoundOver()
      -- only one team remains
      -- PC: changed this call to catch naughty use of GetActiveTeamIndex usage. Use the new 
      -- call to get the last team with a worm thats alive on it.
      SendMessage("WormManager.GetSurvivingTeam")
      WinningTeam = GetData("SurvivingTeamIndex")
      SendIntMessage("GameLogic.Win", WinningTeam)
   else
      StartTurn()
   end
   
end

-- game timer just ran out
function Timer_GameTimedOut()
   -- removed becasue incorrect to end a game mid turn wait until end of turn
   --RoundOver()
   --SendMessage("GameLogic.Draw")
end


function EndTurn()
   -- note: could this happen multiple times per turn (eg from WormDamagedCurrent)?
   -- if so does it matter?

   -- Poisoning now moved to DoPostActivity

   SendMessage("GameLogic.EndTurn")
  
   Activity = GetData("ObjectCount.Active")
   if Activity == 0 then
      SendMessage("Timer.StartPostActivity") -- pause
   else
      WaitUntilNoActivity = true
   end
end


-- If there was some activity its finished now
function GameLogic_NoActivity()
   if WaitUntilNoActivity == true then
      -- reset WaitUntilNoActivity before calling message below because zero time might
      -- cause it to call back immediately and set it true again
      WaitUntilNoActivity = false
      SendMessage("Timer.StartPostActivity")
   end
end



function SetWind()
   SelectRandomWind()
end


function SelectRandomWind()
   -- First the speed
   SendMessage("RandomNumber.Get")
   RandomNumber = GetData("RandomNumber.Float")
   MaxWind = GetData("Wind.MaxSpeed")
   SetData("Wind.Speed", RandomNumber*RandomNumber*MaxWind) -- now squared to make high winds less common

   -- Now a direction
   SendMessage("RandomNumber.Get")
   RandomNumber = GetData("RandomNumber.Float")
   Direction = RandomNumber * 2 * 3.14
   SetData("Wind.Direction", Direction)   
end


function CreateExplosion(DetailObject, WormDamageMag, ImpulseMag, WormDamageRadius, LandDamageRadius, ImpulseRadius)

   SetData("Explosion.DetailObject", DetailObject)
   SetData("Explosion.WormDamageMagnitude", WormDamageMag)
   SetData("Explosion.ImpulseMagnitude", ImpulseMag)
   SetData("Explosion.WormDamageRadius", WormDamageRadius)
   SetData("Explosion.LandDamageRadius", LandDamageRadius)
   SetData("Explosion.ImpulseRadius", ImpulseRadius)
   SendMessage("Explosion.Construct")

end

function SubString( SourceResource, DestResource, SearchString, ReplacementString )
   SetData( "String.SourceDataName", SourceResource )
   SetData( "String.DestDataName", DestResource )
   SetData( "String.SearchString", SearchString )
   SetData( "String.ReplaceString", ReplacementString )
   SendMessage( "String.Substitute" )
end

function SubStringIndirect( SourceResource, DestResource, SearchString, ReplacementResource )
   SetData( "String.SourceDataName", SourceResource )
   SetData( "String.DestDataName", DestResource )
   SetData( "String.SearchString", SearchString )
   SetData( "String.ReplaceString", ReplacementResource )
   SendMessage( "String.SubstituteIndirect" )
end


function GetWormContainerName(Id)

   local DataId = { "Worm.Data00", "Worm.Data01", "Worm.Data02", "Worm.Data03",
                    "Worm.Data04", "Worm.Data05", "Worm.Data06", "Worm.Data07",
                    "Worm.Data08", "Worm.Data09", "Worm.Data10", "Worm.Data11",
                    "Worm.Data12", "Worm.Data13", "Worm.Data14", "Worm.Data15",
                    "Worm.Data16", "Worm.Data17", ""}
  
   -- indexed from 1
  return DataId[Id+1]

end


function GetTeamContainerName(Id)

   local DataId = { "Team.Data00", "Team.Data01", "Team.Data02", "Team.Data03", "" }
  
   -- indexed from 1
  return DataId[Id+1]

end

function GetTeamInventoryName(Id)

   local DataId = { "Inventory.Team00", "Inventory.Team01", "Inventory.Team02", "Inventory.Team03", "" }
  
   -- indexed from 1
  return DataId[Id+1]

end


function GetAllianceInventoryName(Id)

   local DataId = { "Inventory.Alliance00", "Inventory.Alliance01", "Inventory.Alliance02", "Inventory.Alliance03", "" }
  
   -- indexed from 1
  return DataId[Id+1]

end


function GetStockpileInventoryName(Id)

   local DataId = { "Inventory.Stockpile00", "Inventory.Stockpile01", "Inventory.Stockpile02", "Inventory.Stockpile03", "" }
  
   -- indexed from 1
  return DataId[Id+1]

end


function GetWormInventoryName(Id)

   local DataId = { "Inventory.Worm00", "Inventory.Worm01", "Inventory.Worm02", "Inventory.Worm04", 
					"Inventory.Worm05", "Inventory.Worm06", "Inventory.Worm07", "Inventory.Worm08",
					"Inventory.Worm09", "Inventory.Worm10", "Inventory.Worm11", "Inventory.Worm12",
					"Inventory.Worm13", "Inventory.Worm14", "Inventory.Worm15", "Inventory.Worm16",
					 "" }
  
   -- indexed from 1
  return DataId[Id+1]

end

function QueryWormContainer(...)

   if arg.n == 1 then
      Id = arg[1]
   else
      Id = GetData("ActiveWormIndex")
   end

   ContainerName = GetWormContainerName(Id)
   return  QueryContainer(ContainerName) 

end


function DoOncePerTurnFunctions()
end

function RoundOver()
end

function SetWormpotModes()
end
