
-- standard VS game

function Initialise()
   StartedSuddenDeath = false
   
   SetupScheme()
   SetupWormsAndTeams()
 
   SetupMinesAndOildrums() -- do this after worms are set up to get correct worm collision


   local scheme = QueryContainer("GM.SchemeData")
   if scheme.RoundTime == 0 then
      StartSuddenDeath() -- do this after worms are set up to readjust health
   end

   SetData("Camera.StartOfTurnCamera","Default")
   
   SetWormpotModes()

   

   Activity = GetData("ObjectCount.Active")
   if Activity == 0 then
		WaitingForStartFirstTurn = false
        StartFirstTurn()
	else
		WaitingForStartFirstTurn = true
	end
end


function SetupWormsAndTeams()

   WormIndex = 0
   TeamIndex = 0

   GM = QueryContainer("GM.GameInitData")

   if GM.NumberOfTeams>=1 then 
      SetupTeam(GM.T1_Name, GM.T1_NumOfWorms, GM.T1_W1_Name, GM.T1_W2_Name, GM.T1_W3_Name, GM.T1_W4_Name, GM.T1_W5_Name, GM.T1_W6_Name,
               GM.T1_Skill, GM.T1_Grave, GM.T1_Flag, GM.T1_Speech, GM.T1_IsLocal, GM.T1_AlliedGroup)
   end  
   if GM.NumberOfTeams>=2 then 
      SetupTeam(GM.T2_Name, GM.T2_NumOfWorms, GM.T2_W1_Name, GM.T2_W2_Name, GM.T2_W3_Name, GM.T2_W4_Name, GM.T2_W5_Name, GM.T2_W6_Name,
               GM.T2_Skill, GM.T2_Grave, GM.T2_Flag, GM.T2_Speech, GM.T2_IsLocal, GM.T2_AlliedGroup)
   end
   if GM.NumberOfTeams>=3 then 
      SetupTeam(GM.T3_Name, GM.T3_NumOfWorms, GM.T3_W1_Name, GM.T3_W2_Name, GM.T3_W3_Name, GM.T3_W4_Name, GM.T3_W5_Name, GM.T3_W6_Name,
               GM.T3_Skill, GM.T3_Grave, GM.T3_Flag, GM.T3_Speech, GM.T3_IsLocal, GM.T3_AlliedGroup)
   end  
   if GM.NumberOfTeams>=4 then 
      SetupTeam(GM.T4_Name, GM.T4_NumOfWorms, GM.T4_W1_Name, GM.T4_W2_Name, GM.T4_W3_Name, GM.T4_W4_Name, GM.T4_W5_Name, GM.T4_W6_Name,
               GM.T4_Skill, GM.T4_Grave, GM.T4_Flag, GM.T4_Speech, GM.T4_IsLocal, GM.T4_AlliedGroup)
   end


    SendMessage("WormManager.Reinitialise")
end



function SetupTeam(TeamName, NumWorms, Worm1Name, Worm2Name, Worm3Name, Worm4Name, Worm5Name, Worm6Name,
                   Skill, Grave, Flag, Speech, IsLocal, AlliedGroup)

   local lock, team

   lock, team = EditContainer(GetTeamContainerName(TeamIndex)) 
   team.Active = true
   team.Name = TeamName
   team.TeamColour = 0
   team.Skill = Skill
   team.GraveIndex = Grave
   team.FlagGfxName = Flag
   team.AlliedGroup = AlliedGroup
   team.IsLocal = IsLocal
   if Skill>0 then
      team.IsAIControlled = true
   else
      team.IsAIControlled = false
   end

   -- PC: added this to reset the win count for the teams if this is the first round played.
   local iRounds = GetData("GameOver.RoundNumber")
   if iRounds == 0 
   then 
	team.RoundsWon = 0
   end

   CloseContainer(lock) -- must close the container ASAP

   if NumWorms>=1 then 
      SetupWorm(Worm1Name, Speech) 
   end   
   if NumWorms>=2 then 
      SetupWorm(Worm2Name, Speech) 
   end
   if NumWorms>=3 then 
      SetupWorm(Worm3Name, Speech) 
   end
   if NumWorms>=4 then 
      SetupWorm(Worm4Name, Speech) 
   end
   if NumWorms>=5 then 
      SetupWorm(Worm5Name, Speech) 
   end
   if NumWorms>=6 then 
      SetupWorm(Worm6Name, Speech) 
   end


   -- set the inventory
   --CopyContainer("Inventory.Team.Default", GetTeamInventoryName(TeamIndex))

   --SetSpecialWeapon(TeamIndex, SWeapon)
   TeamIndex = TeamIndex+1

end


function SetupWorm(Name, Speech)
   
   local scheme = QueryContainer("GM.SchemeData")
   local lock, worm 

   lock, worm = EditContainer(GetWormContainerName(WormIndex))
   worm.Active = true
   worm.Name = Name
   worm.TeamIndex = TeamIndex
   worm.SfxBankName = Speech    
   worm.Energy = scheme.WormHealth
   worm.WeaponFuse = 3
   worm.WeaponIsBounceMax = false
   worm.Spawn = "spawn"
   CloseContainer(lock)


   WormIndex = WormIndex +1
   
end


function SetSpecialWeapon(TeamIndex, WeaponName)
   SetData("SpecialWeapon.TeamIndex", WormIndex)
   SetData("SpecialWeapon.Weapon", WeaponName)
   SendMessage("GameLogic.SetSpecialWeapon")

   -- TODO
   -- uses GameLogicService to call inventory.SetWeaponCount(0, -1)
end



function DoOncePerTurnFunctions()
   SendMessage("Stats.IgnoreWormDeath") -- no longer credit deaths to current worm
   CheckSuddenDeath()
   SendMessage("GameLogic.DropRandomCrate")
   DoWormpotOncePerTurnFunctions()
end



function SetupScheme()

   local scheme = QueryContainer("GM.SchemeData")

   SetData("GameLogic.ArtilleryMode", scheme.ArtileryMode)
   
   if scheme.FallDamage == 0 then
       -- Don't override the wormpot setting
       local Wormpot = QueryContainer("WormPot")
       if Wormpot.FallingHurtsMore == false then
           SetData("Worm.FallDamageRatio", 0)
      end
   end
   
   SetData("HUD.Clock.DisplayRoundTime", scheme.DisplayTime)
   SetData("Crate.HealthInCrates", scheme.HealthInCrates)
   SetData("DefaultRetreatTime", scheme.LandTime)
   SetData("Land.Indestructable", GetData("FE.Land.Ind"))

   SetupInventoriesAndDelays()
   SetupTeleportIn()
 
   SetData("HotSeatTime", scheme.HotSeat)
   SetData("TurnTime", scheme.TurnTime)
   SetData("RoundTime", scheme.RoundTime)


end



function SetupInventoriesAndDelays()
 
   -- TODO(MS) replace with
   -- local inventory_lock, inventory = EditContainer("Inventory.Team.Default") 
   -- inventory = scheme.Inventory:Clone()
   -- Note: this clone method will not clone containers within this container


   -- Open the containers for read and write
   local inventory_lock, inventory = EditContainer("Inventory.Alliance.Default") 
   local delays_lock, delays = EditContainer("Inventory.WeaponDelays") 
   local scheme = QueryContainer("GM.SchemeData")
 
   -- Airstrike
   delays.Airstrike    = scheme.Airstrike.Delay
   inventory.Airstrike = scheme.Airstrike.Ammo

   -- Armageddon   
   delays.Armageddon = scheme.Armageddon.Delay
   inventory.Grenade = scheme.Armageddon.Ammo

   -- BananaBomb
   delays.BananaBomb    = scheme.BananaBomb.Delay
   inventory.BananaBomb = scheme.BananaBomb.Ammo

   -- BaseballBat
   delays.BaseballBat    = scheme.BaseballBat.Delay
   inventory.BaseballBat = scheme.BaseballBat.Ammo

   -- Bazooka
   delays.Bazooka    = scheme.Bazooka.Delay
   inventory.Bazooka = scheme.Bazooka.Ammo

   -- Bridgekit
   delays.BridgeKit    = scheme.BridgeKit.Delay
   inventory.BridgeKit = scheme.BridgeKit.Ammo

   -- ClusterGrenade
   delays.ClusterGrenade    = scheme.ClusterGrenade.Delay
   inventory.ClusterGrenade = scheme.ClusterGrenade.Ammo

   -- ConcreteDonkey
   delays.ConcreteDonkey    = scheme.ConcreteDonkey.Delay
   inventory.ConcreteDonkey = scheme.ConcreteDonkey.Ammo

   -- CrateShower
   -- NOT IN INVENTORY
   -- delays.CrateShower    = scheme.CrateShower.Delay
   -- inventory.CrateShower = scheme.CrateShower.Ammo

   -- CrateSpy
   -- NOT IN INVENTORY
   -- delays.CrateSpy    = scheme.CrateSpy.Delay
   -- inventory.CrateSpy = scheme.CrateSpy.Ammo

   -- DoubleDamage
   -- NOT IN INVENTORY
   -- delays.DoubleDamage    = scheme.DoubleDamage.Delay
   -- inventory.DoubleDamage = scheme.DoubleDamage.Ammo

   -- DoubleTurnTime
   -- NOT IN INVENTORY
   -- delays.DoubleTurnTime    = scheme.DoubleTurnTime.Delay
   -- inventory.DoubleTurnTime = scheme.DoubleTurnTime.Ammo

   -- Dragonball
   -- DRAGONBALL: delays.Dragonball    = scheme.Dragonball.Delay
   -- DRAGONBALL: inventory.Dragonball = scheme.Dragonball.Ammo

   -- Dynamite
   delays.Dynamite    = scheme.Dynamite.Delay
   inventory.Dynamite = scheme.Dynamite.Ammo

   -- Earthquake
   delays.Earthquake    = scheme.Earthquake.Delay
   inventory.Earthquake = scheme.Earthquake.Ammo

   -- FirePunch
   delays.FirePunch    = scheme.FirePunch.Delay
   inventory.FirePunch = scheme.FirePunch.Ammo

   -- FlameThrower
   -- FLAMETHROWER: delays.FlameThrower    = scheme.FlameThrower.Delay
   -- FLAMETHROWER: inventory.FlameThrower = scheme.FlameThrower.Ammo

   -- Freeze
   delays.Freeze    = scheme.Freeze.Delay
   inventory.Freeze = scheme.Freeze.Ammo
  
   -- GasCanister
   delays.GasCanister    = scheme.GasCanister.Delay
   inventory.GasCanister = scheme.GasCanister.Ammo
 
   -- Girder
   delays.Girder    = scheme.Girder.Delay
   inventory.Girder = scheme.Girder.Ammo
 
   -- Grenade
   delays.Grenade =  scheme.Grenade.Delay
   inventory.Grenade =  scheme.Grenade.Ammo
   
   -- HolyHandGrenade
   delays.HolyHandGrenade    = scheme.HolyHandGrenade.Delay
   inventory.HolyHandGrenade = scheme.HolyHandGrenade.Ammo
 
   -- HomingMissile
   delays.HomingMissile    = scheme.HomingMissile.Delay
   inventory.HomingMissile = scheme.HomingMissile.Ammo
 
   -- HomingPidgeon
   delays.HomingPidgeon    = scheme.HomingPidgeon.Delay
   inventory.HomingPidgeon = scheme.HomingPidgeon.Ammo
 
   -- Jetpack
   delays.Jetpack    = scheme.Jetpack.Delay
   inventory.Jetpack = scheme.Jetpack.Ammo
 
   -- Landmine
   delays.Landmine    = scheme.Landmine.Delay
   inventory.Landmine = scheme.Landmine.Ammo
 
   -- LaserGun
   -- LASER: delays.Lasergun    = scheme.LaserGun.Delay
   -- LASER: inventory.Lasergun = scheme.LaserGun.Ammo
 
   -- LaserSight
   delays.LaserSight    = scheme.LaserSight.Delay
   inventory.LaserSight = scheme.LaserSight.Ammo
 
   -- LaserSight
   delays.LaserSight    = scheme.LaserSight.Delay
   inventory.LaserSight = scheme.LaserSight.Ammo
 
   -- LowGravity
   delays.LowGravity    = scheme.LowGravity.Delay
   inventory.LowGravity = scheme.LowGravity.Ammo

   -- MadCow
   delays.MadCow        = scheme.MadCow.Delay
   inventory.MadCow     = scheme.MadCow.Ammo

   -- MagicBullet
   -- WTF is this here!!
   -- Set to ZERO
   delays.MagicBullet    = 0
   inventory.MagicBullet = 0

   -- MailStrike
   -- MAILSTRIKE: delays.MailStrike    = scheme.MailStrike.Delay
   -- MAILSTRIKE: inventory.MailStrike = scheme.MailStrike.Ammo

   -- MineStrike
   -- WTF is this here!!
   -- Set to ZERO
   delays.MineStrike    = 0
   inventory.MineStrike = 0

   -- MingVase
   -- VASE: delays.MingVase    = scheme.MingVase.Delay
   -- VASE: inventory.MingVase = scheme.MingVase.Ammo

   -- Minigun
   -- MINIGUN: delays.Minigun    = scheme.Minigun.Delay
   -- MINIGUN: inventory.Minigun = scheme.Minigun.Ammo

   -- Mortar
   delays.Mortar    = scheme.Mortar.Delay
   inventory.Mortar = scheme.Mortar.Ammo

   -- Napalm
   -- NAPALM: delays.Napalm    = scheme.Napalm.Delay
   -- NAPALM: inventory.Napalm = scheme.Napalm.Ammo

   -- NinjaRope
   delays.NinjaRope    = scheme.NinjaRope.Delay
   inventory.NinjaRope = scheme.NinjaRope.Ammo

   -- NuclearBomb
   delays.NuclearBomb    = scheme.NuclearBomb.Delay
   inventory.NuclearBomb = scheme.NuclearBomb.Ammo

   -- OldWoman
   delays.OldWoman    = scheme.OldWoman.Delay
   inventory.OldWoman = scheme.OldWoman.Ammo

   -- Parachute
   delays.Parachute    = scheme.Parachute.Delay
   inventory.Parachute = scheme.Parachute.Ammo

   -- PetrolBomb
   delays.PetrolBomb    = scheme.PetrolBomb.Delay
   inventory.PetrolBomb = scheme.PetrolBomb.Ammo

   -- Prod
   delays.Prod    = scheme.Prod.Delay
   inventory.Prod = scheme.Prod.Ammo

   -- QuickWalk
   delays.QuickWalk    = scheme.QuickWalk.Delay
   inventory.QuickWalk = scheme.QuickWalk.Ammo

   -- ScalesOfJustice
   delays.ScalesOfJustice    = scheme.ScalesOfJustice.Delay
   inventory.ScalesOfJustice = scheme.ScalesOfJustice.Ammo

   -- SelectWorm
   delays.ChangeWorm    = scheme.SelectWorm.Delay
   inventory.ChangeWorm = scheme.SelectWorm.Ammo

   -- Sheep
   delays.Sheep    = scheme.Sheep.Delay
   inventory.Sheep = scheme.Sheep.Ammo

   -- SheepStrike
   -- WTF is this here!!
   -- Set to ZERO
   delays.SheepStrike    = 0
   inventory.SheepStrike = 0

   -- Shotgun
   delays.Shotgun    = scheme.Shotgun.Delay
   inventory.Shotgun = scheme.Shotgun.Ammo

   -- SkipGo
   delays.SkipGo     = 0
   inventory.SkipGo  = -1

   -- Surrender
   delays.Surrender     = 0
   inventory.Surrender  = -1

   -- SuperBananaBomb
   -- SUPERBANANBOMB: delays.SuperBananaBomb    = scheme.SuperBananaBomb.Delay
   -- SUPERBANANBOMB: inventory.SuperBananaBomb = scheme.SuperBananaBomb.Ammo

   -- SuperSheep
   delays.SuperSheep    = scheme.SuperSheep.Delay
   inventory.SuperSheep = scheme.SuperSheep.Ammo

   -- Teleport
   delays.Teleport    = scheme.Teleport.Delay
   inventory.Teleport = scheme.Teleport.Ammo

   -- Uzi
   delays.Uzi    = scheme.Uzi.Delay
   inventory.Uzi = scheme.Uzi.Ammo

   -- VikingAxe
   delays.VikingAxe    = scheme.VikingAxe.Delay
   inventory.VikingAxe = scheme.VikingAxe.Ammo

   -- Lottery Strike
   delays.LotteryStrike    = scheme.LotteryStrike.Delay
   inventory.LotteryStrike = scheme.LotteryStrike.Ammo
   
   -- DoctorsStrike
   delays.DoctorsStrike    = scheme.DoctorsStrike.Delay
   inventory.DoctorsStrike = scheme.DoctorsStrike.Ammo
   
   -- MegaMine
   delays.MegaMine    = scheme.MegaMine.Delay
   inventory.MegaMine = scheme.MegaMine.Ammo
   
   -- StickyBomb
   delays.StickyBomb    = scheme.StickyBomb.Delay
   inventory.StickyBomb = scheme.StickyBomb.Ammo
   
   -- Binoculars
   delays.Binoculars    = scheme.Binoculars.Delay
   inventory.Binoculars = scheme.Binoculars.Ammo
   
   -- Blowpipe
   delays.Blowpipe    = scheme.Blowpipe.Delay
   inventory.Blowpipe = scheme.Blowpipe.Ammo

   -- Red bull
   delays.Redbull     = scheme.Redbull.Delay
   inventory.Redbull  = scheme.Redbull.Ammo

   -- Close the lock
   CloseContainer(delays_lock)
   CloseContainer(inventory_lock)


   -- If this is the first round then no matter what the stockpiling options are
   -- of if "Stockpiling Off" then give every Alliance a copy of the Scheme inventory
   local RoundsCompleted = GetData("GameOver.RoundNumber")
   if RoundsCompleted == 0 or scheme.Stockpiling == 0 then
      CopyContainer("Inventory.Alliance.Default", "Inventory.Alliance00")
      CopyContainer("Inventory.Alliance.Default", "Inventory.Alliance01")
      CopyContainer("Inventory.Alliance.Default", "Inventory.Alliance02")
      CopyContainer("Inventory.Alliance.Default", "Inventory.Alliance03")
   -- "Stockpiling" this option gives the alliance the total of the default and the stockpile
   elseif scheme.Stockpiling == 1 then
      for i = 0,3 do
         InventoryName = GetAllianceInventoryName(i)
         StockpileName = GetStockpileInventoryName(i)
         CopyContainer("Inventory.Alliance.Default", InventoryName)
         SetData("GameLogic.AddInventory.Arg0", InventoryName)
         SetData("GameLogic.AddInventory.Arg1", StockpileName)
         SendMessage("GameLogic.AddInventory")
      end
   -- "Stockpiling Anti" (Stock continues to decrement throught all turns)
   elseif scheme.Stockpiling == 2 then 
      CopyContainer("Inventory.Stockpile00", "Inventory.Alliance00")
      CopyContainer("Inventory.Stockpile01", "Inventory.Alliance01")
      CopyContainer("Inventory.Stockpile02", "Inventory.Alliance02")
      CopyContainer("Inventory.Stockpile03", "Inventory.Alliance03")
   end

   -- Special weapons
   -- if option is enabled then the Alliance inventories are supplemented with
   -- the combined special weapons of all the teams
   GM = QueryContainer("GM.GameInitData")
   if scheme.Special == 1 then
       -- only add in the specials when we add the full inventories
       if RoundsCompleted == 0 or scheme.Stockpiling == 0 or scheme.Stockpiling == 1 then
          if GM.NumberOfTeams>=1 then 
             IncrementAlliedInventory(GM.T1_AlliedGroup, GM.T1_SWeapon)
          end
          if GM.NumberOfTeams>=2 then 
             IncrementAlliedInventory(GM.T2_AlliedGroup, GM.T2_SWeapon)
          end
          if GM.NumberOfTeams>=3 then 
             IncrementAlliedInventory(GM.T3_AlliedGroup, GM.T3_SWeapon)
          end
          if GM.NumberOfTeams>=4 then 
             IncrementAlliedInventory(GM.T4_AlliedGroup, GM.T4_SWeapon)
          end
       end
   end
   
end


function IncrementAlliedInventory(AlliedGroup, Weapon)
   local InventoryName = GetAllianceInventoryName(AlliedGroup)
   SetData("GameLogic.AddInventory.Arg0", InventoryName)
   SetData("GameLogic.AddInventory.Arg1", Weapon)
   SendMessage("GameLogic.IncrementInventory")
end


function Stockpile()
   -- record the remaining stock for future rounds
   CopyContainer("Inventory.Alliance00", "Inventory.Stockpile00")
   CopyContainer("Inventory.Alliance01", "Inventory.Stockpile01")
   CopyContainer("Inventory.Alliance02", "Inventory.Stockpile02")
   CopyContainer("Inventory.Alliance03", "Inventory.Stockpile03")
end



function RoundOver()
   Stockpile()
end

-- fudge together a function to add a worms stats onto its team stats
function AddWormStatsToTeamStats(index,ws,team1,team2,team3,team4)
	-- find out which team this worm is on
    local DataId = GetWormContainerName(index)
    local worm = QueryContainer(DataId)
	iteam = worm.TeamIndex
	-- given the team, add the worm stats to the team stats
	local ts
	if iteam == 0 then ts = team1 end
	if iteam == 1 then ts = team2 end
	if iteam == 2 then ts = team3 end
	if iteam == 3 then ts = team4 end
	
	-- now add them together
	ts.PositiveDamage = ts.PositiveDamage + ws.PositiveDamage
	ts.PositiveKills = ts.PositiveKills + ws.PositiveKills
	ts.ShotsFired = ts.ShotsFired + ws.ShotsFired
	ts.ShotsMissed = ts.ShotsMissed + ws.ShotsMissed
	ts.SurvivalTime = ts.SurvivalTime + ws.SurvivalTime
	ts.StrategyPoints = ts.StrategyPoints + ws.StrategyPoints
	ts.BoringPoints = ts.BoringPoints + ws.BoringPoints
	ts.NegativeDamage = ts.NegativeDamage + ws.NegativeDamage
	ts.CratesCollected = ts.CratesCollected + ws.CratesCollected
	ts.RopesUsed = ts.RopesUsed + ws.RopesUsed
	ts.BestShotDamage = ts.BestShotDamage + ws.BestShotDamage
	ts.BestShotKills = ts.BestShotKills + ws.BestShotKills

end

function GetWormTeamIndex(index)
    local DataId = GetWormContainerName(index)
    local worm = QueryContainer(DataId)
	return worm.TeamIndex
end

function GetWormName(index)
if index < 0 then return "" end
if index > 16 then return "" end
    local DataId = GetWormContainerName(index)
    local worm = QueryContainer(DataId)
	return worm.Name
end

function GetTeamName(index)
if index < 0 then return "" end
if index > 4 then return "" end
    local DataId = GetTeamContainerName(index)
    local team = QueryContainer(DataId)
	return team.Name
end
-- this function is the place where you actually calculate the stats..
function Stats_CalculateGameStats()

	SetData("GameStats.StatString","")
	g_bSentWormStat = false
	g_bSentTeamStat = false
	
	-- load the worm stats containers.
	local numworms
	local wotmmax				= { 0, -1, 0 }
	local bestshotmax			= { 0, -1 }
	local mostboringmax			= { 0, -1 }
	local mostinterestingmax	= { 0, -1 }
	local greediestmax			= { 0, -1 }
	local strategicmax			= { 0, -1 }
	local uselessmin			= { 0, -1 }
	local accuracymax			= { 0, -1 }
	local leastpopularmax		= { 0, -1 }
	local survivalmax  		    = { 0, -1 }
	local entertainmentmax  	= { 0, -1 }
	local shymin				= { 1000000, -1 }
	
	local teamboringmax			= { 0, -1 }
	local teamexcitingmax		= { 0, -1 }
	local teamgreedymax			= { 0, -1 }
	local teamstrategicmax		= { 0, -1 }
	local teamuselessmax		= { 0, -1 }
	local teamaccuratemax		= { 0, -1 }
	local teamentertainingmax	= { 0, -1 }
	
	-- open the team stats containers for edit, so we can accumulate the stats
	local team1_name = GetTeamStatsContainer(0)
	local team1_lock,team1 = EditContainer(team1_name)
	ClearStats(team1)
	local team2_name = GetTeamStatsContainer(1)
	local team2_lock,team2 = EditContainer(team2_name)
	ClearStats(team2)
	local team3_name = GetTeamStatsContainer(2)
	local team3_lock,team3 = EditContainer(team3_name)
	ClearStats(team3)
	local team4_name = GetTeamStatsContainer(3)
	local team4_lock,team4 = EditContainer(team4_name)
	ClearStats(team4)
	
	-- is this 0..15 ?
	for numworms = 0,15 do
		-- fudge all game stats data items into local variables
		local wormname = GetWormStatsContainer(numworms)
		local ws = QueryContainer(wormname)
		local a = ws.PositiveDamage
		local b = ws.PositiveKills
		local c = ws.ShotsFired
		local d = ws.ShotsMissed
		local e = ws.SurvivalTime
		local f = ws.StrategyPoints
		local g = ws.BoringPoints
		local h = ws.NegativeDamage
		local i = ws.CratesCollected
		local j = ws.RopesUsed
		local k = ws.BestShotDamage
		local l = ws.BestShotKills
		
		-- accumulate the team stats for the given worm.
		AddWormStatsToTeamStats(numworms,ws,team1,team2,team3,team4)
		
		-- actually calculate this worms stats.... add what you want here.
		
		local wotm =  (( ( a - h ) / c ) * b )
		local bestshot = ( l * l ) * k
		local mostboring = g + d
		local mostinteresting = (a + h) * b - (( g + d ) * 10 )
		local greediest = i
		local strategic = f
		local useless = e-h-d
		local accuracy = a/c
		local leastpopular = h
		local survival = e
		local entertainment = ( ( a * b ) * i) 
		local shy = e
		
		-- now work out wether this guy has higher or lower stats for a given value.
		if wotm > wotmmax[1] then 
			wotmmax[3] = b	
			wotmmax[2] = numworms 
			wotmmax[1] = wotm	
		end	
		if bestshot > bestshotmax[1] then 
			bestshotmax[2] = numworms
			bestshotmax[1] = bestshot
		end	
		if mostboring > mostboringmax[1] then 
			mostboringmax[2] = numworms
			mostboringmax[1] = mostboring
		end	
		if mostinteresting > mostinterestingmax[1] then 
			mostinterestingmax[2] = numworms
			mostinterestingmax[1] = mostinteresting
		end
		if greediest > greediestmax[1] then 
			greediestmax[2] = numworms
			greediestmax[1] = greediest
		end
		if strategic > strategicmax[1] then 
			strategicmax[2] = numworms
			strategicmax[1] = strategic
		end	
		if useless > uselessmin[1] then 
			uselessmin[2] = numworms
			uselessmin[1] = useless
		end	
		if accuracy > accuracymax[1] then 
			accuracymax[2] = numworms
			accuracymax[1] = accuracy
		end	
		if leastpopular > leastpopularmax[1] then 
			leastpopularmax[2] = numworms
			leastpopularmax[1] = leastpopular
		end	
		if survival > survivalmax[1] then 
			survivalmax[2] = numworms
			survivalmax[1] = survival
		end	
		if entertainment > entertainmentmax[1] then 
			entertainmentmax[2] = numworms
			entertainmentmax[1] = entertainment
		end	
		if shy ~= 0 and shy < shymin[1] then 
			shymin[2] = numworms
			shymin[1] = shy
		end	
		
	end
	-- calculate team stats

	local teamnum
	for teamnum = 0,3 do
		if teamnum == 0 then ts = team1 end
		if teamnum == 1 then ts = team2 end
		if teamnum == 2 then ts = team3 end
		if teamnum == 3 then ts = team4 end
		
		local a = ts.PositiveDamage
		local b = ts.PositiveKills
		local c = ts.ShotsFired
		local d = ts.ShotsMissed
		local e = ts.SurvivalTime
		local f = ts.StrategyPoints
		local g = ts.BoringPoints
		local h = ts.NegativeDamage
		local i = ts.CratesCollected
		local j = ts.RopesUsed
		local k = ts.BestShotDamage
		local l = ts.BestShotKills
			
		local boringteam = g + d
		local excitingteam = (a + h) * b - (( g + d ) * 10 )
		local greedyteam = i
		local strategicteam = f
		local uselessteam = e
		local accurateteam = a/c
		local entertainingteam = ((a + b) * i) - (( g + d ) * 10 )	
		
		if boringteam > teamboringmax[1] then 
			teamboringmax[1] = boringteam
			teamboringmax[2] = teamnum
		end
		if excitingteam > teamexcitingmax[1] then
			teamexcitingmax[1] = excitingteam
			teamexcitingmax[2] = teamnum
				
		end
		if greedyteam > teamgreedymax[1] then
			teamgreedymax[1] = greedyteam
			teamgreedymax[2] = teamnum
		end
		if strategicteam > teamstrategicmax[1] then
			teamstrategicmax[1] = strategicteam
			teamstrategicmax[2] = teamnum
		end
		if uselessteam > teamuselessmax[1] then
			teamuselessmax[1] = uselessteam
			teamuselessmax[2] = teamnum
		end
		if accurateteam > teamaccuratemax[1] then
			teamaccuratemax[1] = accurateteam
			teamaccuratemax[2] = teamnum
		end
		if entertainingteam > teamentertainingmax[1] then
			teamentertainingmax[1] = entertainingteam
			teamentertainingmax[1] = teamnum
		end
		
	end		

    --The Worm of the Match Award goes to %s (%s) with %d kill(s)
	if wotmmax[1] > 15 then
		SendWormAndTeamStat(wotmmax[2],"FE.Award.WOTM",wotmmax[3])
	end
	
        
	--%s wins the sniping award...
	if bestshotmax[1] > 80 then
		SendWormStat(bestshotmax[2],"FE.Award.Sniping")
	end
	
        
	--Most boring worm
	if mostboringmax[1] > 3 then
		SendWormStat(mostboringmax[2],"FE.Award.Boring")
	end
	
	--Crowd Pleaser Award goes to %s.
	if mostinterestingmax[1] > 80 then
		SendWormStat(mostinterestingmax[2],"FE.Award.Interesting")
	end
	
	--The Happy Shopper Award was collected by %s.
	if greediestmax[1] > 3 then
		SendWormStat(greediestmax[2],"FE.Award.Greedy")
	end
	
	--Most strategic worm
	if strategicmax[1] > 6 then
		SendWormStat(strategicmax[2],"FE.Award.Strategic")
	end
	
        
	--%s (%s) was a liability.
	if uselessmin[1] > 1800000 then
		--SendWormStat(uselessmin[2],"FE.Award.Liability")
		SendWormAndTeamStat(uselessmin[2],"FE.Award.Liability",uselessmin[1])
	end
	
        
	--The accuracy award was handed to %s
	if accuracymax[1] > 35 then
		SendWormStat(accuracymax[2],"FE.Award.accuracy")
	end
	
	-- %s had nothing to say.
	if leastpopularmax[1] > 10 then
		SendWormStat(leastpopularmax[2],"FE.Award.Outcast")
	end
	
	--Survival Award
	if survivalmax[1] > 2000000 then
		SendWormStat(survivalmax[2],"FE.Award.Survival")
	end
	
        
	--Crowd Pleaser Award goes to %s.	
	if entertainmentmax[1] > 170 then
		SendWormStat(entertainmentmax[2],"FE.Award.Entertainment")
	end
        
	StatPrint(" \n")

	--%s were the most boring team.
	if teamboringmax[1] > 5 then
	    SendTeamStat(teamboringmax[2],"FE.TeamAward.Boring")
	end
   
	--Most exciting team
	if teamexcitingmax[1] > 400 then
		SendTeamStat(teamexcitingmax[2],"FE.TeamAward.Exciting")
	end
       
        
	--The Shopping Award goes to %s.
	if teamgreedymax[1] > 5 then
		SendTeamStat(teamgreedymax[2],"FE.TeamAward.Greedy")
	end
                
	--Most strategic team
	if teamstrategicmax[1] > 8 then
		SendTeamStat(teamstrategicmax[2],"FE.TeamAward.Strategic")
	end
        
	--%s were not very good at all.
	if teamuselessmax[1] > 2600000 then
		SendTeamStat(teamuselessmax[2],"FE.TeamAward.Poor")
	end
        
	--Most accurate team
	if teamaccuratemax[1] > 15 then
		SendTeamStat(teamaccuratemax[2],"FE.TeamAward.Accurate")
	end
        
	--Most entertaining team
        if teamentertainingmax[1] > 3 then
		SendTeamStat(teamentertainingmax[2],"FE.TeamAward.Entertaining")
	end
	
	
	-- close the team containers
	CloseContainer(team1_lock)
	CloseContainer(team2_lock)
	CloseContainer(team3_lock)
	CloseContainer(team4_lock)
	
end

function DebugStats(string,value,index)
	name = GetWormName(index)
	OutputDebugStats(string,value,index,name)
end

function DebugTeamStats(string,value,index)
	name = GetTeamName(index)
	OutputDebugStats(string,value,index,name)
end

function OutputDebugStats(string,value,index,name)
	SetData("Stats.SourceString","STR : %d : %i : %s")		
	SubString("Stats.SourceString","Stats.DestString","STR",string)	
	copystring = GetData("Stats.DestString")
	SetData("Stats.SourceString",copystring)			
	SetData("Stats.WormName",name)
	SetData("Stats.Value",value)
	SetData("Stats.Index",index)
	SendMessage("GameStats.Print")
end

function SendWormStat(index,statresource)
	if g_bSentWormStat ==  false then
		local wormstring = GetData("FE.Tip.Worm")
		StatPrint(wormstring)
		StatPrint("-----------------")
		g_bSentWormStat = true
	end
	name = GetWormName(index)	
	string = GetData(statresource)
	SendStat(name,string)
end

function SendTeamStat(index,statresource)
	if g_bSentTeamStat ==  false then
		local teamstring = GetData("FE.Tip.Team ")
		StatPrint(teamstring)
		StatPrint("-----------------")
		g_bSentTeamStat = true
	end
	name = GetTeamName(index)
	string = GetData(statresource)
	SendStat(name,string)
end

function SendWormAndTeamStat(wormindex,statresource,value)
	if g_bSentWormStat ==  false then
		local tipstring = GetData("FE.Tip.Worm")
		StatPrint(tipstring)
		StatPrint("-----------------")
		g_bSentWormStat = true
	end
	wormname = GetWormName(wormindex)
	teamindex = GetWormTeamIndex(wormindex)
	teamname = GetTeamName(teamindex)
	string = GetData(statresource)
	SetData("Stats.SourceString",string)	
	SetData("Stats.TeamName",teamname)
	SetData("Stats.WormName",wormname)
	SetData("Stats.Value",value)
	SendMessage("GameStats.Print")
end

function StatPrint(string)
	SetData("Stats.DestString",string)
	SendMessage("GameStats.AddStatText")
end

function SendStat(name,string)
--if name == "" then return end
	SetData("Stats.SourceString",string)
	SubString("Stats.SourceString","Stats.DestString","%s",name)
	SendMessage("GameStats.AddStatText")
end
function ClearStats(ws)
	ws.PositiveDamage = 0
	ws.PositiveKills = 0
	ws.ShotsFired = 0
	ws.ShotsMissed = 0
	ws.SurvivalTime = 0
	ws.StrategyPoints = 0
	ws.BoringPoints = 0
	ws.NegativeDamage = 0
	ws.CratesCollected = 0
	ws.RopesUsed = 0
	ws.BestShotDamage = 0
	ws.BestShotKills = 0
end
-- function to return a name given a worm index
function GetWormStatsContainer(index)
   local DataId = { "GameStats.Worm00", "GameStats.Worm01", "GameStats.Worm02", "GameStats.Worm03",
                    "GameStats.Worm04", "GameStats.Worm05", "GameStats.Worm06", "GameStats.Worm07",
                    "GameStats.Worm08", "GameStats.Worm09", "GameStats.Worm10", "GameStats.Worm11",
                    "GameStats.Worm12", "GameStats.Worm13", "GameStats.Worm14", "GameStats.Worm15",
                    "GameStats.Worm16", "GameStats.Worm17", ""}
  
   -- indexed from 1
  return DataId[index + 1]
end

-- function to return a name given a worm index
function GetTeamStatsContainer(index)
   local DataId = { "GameStats.Team00", "GameStats.Team01", "GameStats.Team02", "GameStats.Team03",
                    "GameStats.Team04", "GameStats.Team05", ""}
  
   -- indexed from 1
  return DataId[index + 1]
end


function UpdateTurnStats()
   -- calculate all stats for the current turn before we carry on to the victory conditions etc.
   SendMessage("Stats.UpdateAtEndOfTurn")
end

-- overridden so that we can process some stats at the end of each turn
function TurnEnded()
   UpdateTurnStats()
   CheckOneTeamVictory()      
end



function TurnStarted()
   local scheme = QueryContainer("GM.SchemeData")
   if scheme.WormSelect == 1 then
      SendMessage("WormSelect.OptionSelected")
   end
   TeleportIn()
end



function SetupTeleportIn()
   -- Must do this BEFORE call to ReinitialiseWorms
   local scheme = QueryContainer("GM.SchemeData")
   local WormIndex = 0
   local DataId = GetWormContainerName(WormIndex)
   while DataId ~= "" do
      
      lock, worm = EditContainer(DataId)
      if scheme.TeleportIn == 1 then
         worm.TeleportIn = true
      else
         worm.TeleportIn = false
      end
      CloseContainer(lock)
      
      WormIndex = WormIndex + 1
      DataId = GetWormContainerName(WormIndex)
   end
end


function TeleportIn()
   local CurrentWorm = GetData("ActiveWormIndex")
   if CurrentWorm ~= -1 then
      local DataId = GetWormContainerName(CurrentWorm)
      local worm = QueryContainer(DataId)
      if worm.TeleportIn == true then
         SendMessage("WormManager.TeleportIn")
      end
   end
end


function SetupMinesAndOildrums()
   
   -- note do this AFTER worms have been spawned in so they apear in the collision tests
   

   local scheme = QueryContainer("GM.SchemeData")
   
   --SetData("Mine.DudProbabilty", 0.1)
   
   if scheme.MineFuse == 0 then
      -- instant explosions
      SetData("Mine.MinFuse", 0)
      SetData("Mine.MaxFuse", 0)
   else
      -- random
      SetData("Mine.MinFuse", 0)
      SetData("Mine.MaxFuse", 3000)
   end

   -- 0 = None, 1 = mines, 2 = drums, 3 = both
   local i
   if scheme.Objects == 1 or scheme.Objects == 3 then
      for i = 1,10 do
         SendMessage("GameLogic.CreateRandomMine")
      end
   end

   if scheme.Objects == 2 or scheme.Objects == 3 then
      for i = 1,10 do
         SendMessage("GameLogic.CreateRandomOildrum")
      end
   end

end


function Timer_GameTimedOut()
   -- game over dealth with in CheckSuddenDeath()
end


function CheckSuddenDeath()

   -- dont start or continue sudden death if a win is pending
   -- ie. if a team already won dont risk killing it due to sudden death
   SendMessage("WormManager.GetActiveAlliances")
   local AllianceCount = GetData("AllianceCount")

   if AllianceCount>1 then
      local RoundTimeRemaining = GetData("RoundTimeRemaining")
   
      if RoundTimeRemaining == 0 and StartedSuddenDeath==false then
         StartSuddenDeath()
      end

      -- Raise the water level
      SetData("Water.Level", GetData("Water.Level") + GetData("Water.RiseSpeed.Current") )
   end
end


function StartSuddenDeath()
   -- called at start if No game timer

   -- Set Sudden Death Options
   -- (0 = Nuke, 1 = One health, 2 = Water rise, 3 = GameOver)

   StartedSuddenDeath = true

   local scheme = QueryContainer("GM.SchemeData")
   if scheme.SuddenDeath == 0 then
      SendMessage("Comment.SuddenDeath")
      SendMessage("GameLogic.CreateNuclearEffect")
   elseif scheme.SuddenDeath == 1 then
      SendMessage("Comment.SuddenDeath")
      SetAllWormsEnergy(1)
   -- unecessary (we can switch it on and off by the later level rise button)
   --   elseif scheme.SuddenDeath == 2 then
   elseif scheme.SuddenDeath == 3 then
      RoundOver()
      SendMessage("GameLogic.Draw")
   end

   -- Set Water Rise Speed
   -- (0 = No Rise, 1 = Slow rise, 2 = Medium rise, 3 = fast rise)
   if scheme.WaterSpeed == 0 then
      SetData("Water.RiseSpeed.Current", 0)
   elseif scheme.WaterSpeed == 1 then
      SetData("Water.RiseSpeed.Current", GetData("Water.RiseSpeed.Slow") )
   elseif scheme.WaterSpeed == 2 then
      SetData("Water.RiseSpeed.Current", GetData("Water.RiseSpeed.Medium") )
   elseif scheme.WaterSpeed == 3 then
      SetData("Water.RiseSpeed.Current", GetData("Water.RiseSpeed.Fast") )
   end


end


function SetAllWormsEnergy(energy)
   local WormIndex = 0
   local DataId = GetWormContainerName(WormIndex)
   while DataId ~= "" do
      lock, worm = EditContainer(DataId)
      worm.Energy = energy
      CloseContainer(lock)
      WormIndex = WormIndex + 1
      DataId = GetWormContainerName(WormIndex)
   end
end

-- If there was some activity its finished now
function GameLogic_NoActivity()
   if WaitUntilNoActivity == true then
      SendMessage("Timer.StartPostActivity")
      WaitUntilNoActivity = false
   end
   
   if WaitingForStartFirstTurn == true then
      StartFirstTurn()
      WaitingForStartFirstTurn = false
   end
   
end


function Worm_Died()
   -- if there is only one team left abort the turn now rather than wait on
   SendMessage("WormManager.GetActiveAlliances")
   AllianceCount = GetData("AllianceCount")
   if AllianceCount < 2 then
      SendMessage("Weapon.Delete")
      SendMessage("Timer.EndTurn") 
      SendMessage("Weapon.DisableWeaponChange")
   end
end
